from typing import Callable, List, Dict, Tuple

from tunepy2.interfaces import AbstractGenomeFactory, AbstractRandomNumberGenerator, AbstractGenomeComparer
from tunepy2.internal import Genome, DimensionsMismatchException


class UniformCrossoverGenomeFactory(AbstractGenomeFactory):
    """ Genome factory that allows crossover of bits from either parent """

    def __init__(self,
                 dimensions: Tuple,
                 rng: AbstractRandomNumberGenerator,
                 mutation_rate: float,
                 comparer: AbstractGenomeComparer,
                 fitness_func: Callable[..., float],
                 *args,
                 **kwargs):
        """ Creates a new UniformCrossoverGenomeFactory

        :param dimensions: dimensions of bitstrings of Genome objects generated by the factory
        :param rng: random number generator
        :param mutation_rate: probability of mutation (must be (0.0, 1.0))
        :param comparer: comparer used to select Genome objects
        :param fitness_func: fitness function passed into new Genome objects that accepts a bitstring
        :param args: will be passed into fitness_func
        :param kwargs: will be passed into fitness_func
        """
        self._dimensions = dimensions
        self._rng = rng
        self._mutation_rate = mutation_rate
        self._comparer = comparer
        self._fitness_func = fitness_func
        self._args = args
        self._kwargs = kwargs

    def build(self, prior_genomes: List[Genome]) -> Genome:
        """ Builds a new Genome from a list of priors

        :param prior_genomes: list of prior Genomes
        :return: a new Genome object
        """
        def recursive_gene_selection(sub_dimensions, sub_new_bitstring, sub_left_bitstring, sub_right_bitstring):
            if not (sub_dimensions[0] == len(sub_right_bitstring) and
                    sub_dimensions[0] == len(sub_left_bitstring)):
                raise DimensionsMismatchException

            if len(sub_dimensions) == 1:
                for index in range(sub_dimensions[0]):
                    if self._rng.random() > self._mutation_rate:
                        if self._rng.random() < 0.5:
                            sub_new_bitstring[index] = sub_left_bitstring[index]
                        else:
                            sub_new_bitstring[index] = sub_right_bitstring[index]
            else:
                for index in range(sub_dimensions[0]):
                    recursive_gene_selection(
                        sub_dimensions[1:],
                        sub_new_bitstring[index],
                        sub_left_bitstring[index],
                        sub_right_bitstring[index])

        new_bitstring = self._rng.random_int_array(0, 2, self._dimensions)
        left_prior = self._comparer.compare(prior_genomes).bitstring
        right_prior = self._comparer.compare(prior_genomes).bitstring

        recursive_gene_selection(self._dimensions, new_bitstring, left_prior, right_prior)

        return Genome(self._fitness_func, new_bitstring, *self._args, **self._kwargs)

    @property
    def dimensions(self) -> Tuple:
        """

        :return: the dimensions of Genome objects returned by this factory
        """
        return self._dimensions


class SinglePointCrossoverGenomeFactory(AbstractGenomeFactory):
    """ Genome factory that takes all bits prior to a crossover point from one parent and all bits after and
    including a crossover point from the other parent """

    def __init__(self,
                 dimensions: Tuple,
                 rng: AbstractRandomNumberGenerator,
                 mutation_rate: float,
                 comparer: AbstractGenomeComparer,
                 fitness_func: Callable[..., float],
                 *args,
                 **kwargs):
        """ Creates a new SinglePointCrossoverGenomeFactory

        :param dimensions: dimensions of bitstrings of Genome objects generated by the factory
        :param rng: random number generator
        :param mutation_rate: probability of mutation (must be (0.0, 1.0))
        :param comparer: comparer used to select Genome objects
        :param fitness_func: fitness function passed into new Genome objects that accepts a bitstring
        :param args: will be passed into fitness_func
        :param kwargs: will be passed into fitness_func
        """
        self._dimensions = dimensions
        self._rng = rng
        self._mutation_rate = mutation_rate
        self._comparer = comparer
        self._fitness_func = fitness_func
        self._args = args
        self._kwargs = kwargs

    def build(self, prior_genomes: List[Genome]) -> Genome:
        """ Builds a new Genome from a list of priors

        :param prior_genomes: list of prior Genomes
        :return: a new Genome object
        """
        def recursive_gene_selection(sub_dimensions, sub_new_bitstring, sub_left_bitstring, sub_right_bitstring):
            if not (sub_dimensions[0] == len(sub_right_bitstring) and
                    sub_dimensions[0] == len(sub_left_bitstring)):
                raise DimensionsMismatchException

            if len(sub_dimensions) == 1:
                sub_new_bitstring[:crossover_index] = sub_left_bitstring[:crossover_index]
                sub_new_bitstring[crossover_index:] = sub_right_bitstring[crossover_index:]

                for index in range(sub_dimensions[0]):
                    if self._rng.random() < self._mutation_rate:
                        sub_new_bitstring[index] = self._rng.random_int(0, 2)
            else:
                for index in range(sub_dimensions[0]):
                    recursive_gene_selection(
                        sub_dimensions[1:],
                        sub_new_bitstring[index],
                        sub_left_bitstring[index],
                        sub_right_bitstring[index])

        new_bitstring = self._rng.random_int_array(0, 2, self._dimensions)
        left_prior = self._comparer.compare(prior_genomes).bitstring
        right_prior = self._comparer.compare(prior_genomes).bitstring
        crossover_index = self._rng.random_int(1, self._dimensions[-1])

        recursive_gene_selection(self._dimensions, new_bitstring, left_prior, right_prior)

        return Genome(self._fitness_func, new_bitstring, *self._args, **self._kwargs)

    @property
    def dimensions(self) -> Tuple:
        """

        :return: the dimensions of Genome objects returned by this factory
        """
        return self._dimensions



from tunepy2 import Genome
from tunepy2.interfaces import AbstractOptimizer, AbstractOptimizerBuilder, AbstractGenomeFactory, \
    AbstractConvergenceCriterion


class BasicRestartOptimizer(AbstractOptimizer):
    """
    An optimizer that creates other optimizers, tests them to convergence, and repeats until its own convergence
    criterion is met.
    """
    def __init__(
            self,
            optimizer_builder: AbstractOptimizerBuilder,
            genome_factory: AbstractGenomeFactory,
            population_size: int,
            convergence_criterion: AbstractConvergenceCriterion):
        """
        Creates a new BasicRestartOptimizer.

        :param optimizer_builder: builds new optimizers for each iteration of this optimizer
        :param genome_factory: creates Genome objects
        :param population_size: population size for each underlying optimizer
        :param convergence_criterion: convergence criterion used for evaluation of this optimizer
        """
        self._optimizer_builder = optimizer_builder
        self._genome_factory = genome_factory
        self._population_size = population_size
        self._convergence_criterion = convergence_criterion
        self._max_fitness = float('-inf')
        self._converged = False
        self._best_genome = None

    def next(self):
        """
        Performs the next iteration of optimization.

        """
        new_optimizer = self._optimizer_builder\
            .clear()\
            .add_to_initial_population_from_factory(self._genome_factory, self._population_size)\
            .build()

        while not new_optimizer.converged:
            new_optimizer.next()

        self._converged = self._convergence_criterion.converged(self.best_genome, new_optimizer.best_genome)

        if new_optimizer.best_genome.fitness > self._max_fitness:
            self._max_fitness = new_optimizer.best_genome.fitness
            self._best_genome = new_optimizer.best_genome

    @property
    def converged(self) -> bool:
        """
        Whether or not this algorithm has converged

        :return: true when this algorithm has converged or false if not
        """
        return self._converged

    @property
    def best_genome(self) -> Genome:
        """
        The best genome so far

        :return: a Genome instance
        """
        return self._best_genome



import unittest

from tunepy2.optimizers import BasicAnnealingOptimizer
from tunepy2.optimizers.builders import BasicAnnealerOptimizerBuilder
from tunepy2.interfaces.stubs import PassThroughAnnealingSchedule
from tunepy2.interfaces.stubs import PassThroughGenomeFactory
from tunepy2.random import NumpyRNG
from tunepy2 import Genome, InitialPopulationUndefinedException


class TestOptimizerBasicBuilder(unittest.TestCase):
    def test_add_to_initial_population_from_factory_returns_self(self):
        def fitness_func(bitstring):
            self.fail()
            return 0.0

        genome_factory = PassThroughGenomeFactory(Genome.new_default_genome((5,), fitness_func))
        annealing_schedule = PassThroughAnnealingSchedule(0, True)
        optimizer_builder = BasicAnnealerOptimizerBuilder((5,),
                                                          NumpyRNG(),
                                                          genome_factory,
                                                          annealing_schedule,
                                                          fitness_func)

        returned_object = optimizer_builder.add_to_initial_population_from_factory(genome_factory, 1)
        self.assertEqual(optimizer_builder, returned_object)

    def test_add_to_initial_population_returns_self(self):
        def fitness_func(bitstring):
            self.fail()
            return 0.0

        genome_factory = PassThroughGenomeFactory(Genome.new_default_genome((5,), fitness_func))
        annealing_schedule = PassThroughAnnealingSchedule(0, True)
        optimizer_builder = BasicAnnealerOptimizerBuilder((5,),
                                                          NumpyRNG(),
                                                          genome_factory,
                                                          annealing_schedule,
                                                          fitness_func)

        returned_object = optimizer_builder.add_to_initial_population(Genome.new_default_genome((5,), fitness_func))
        self.assertEqual(optimizer_builder, returned_object)

    def test_add_to_initial_population_from_factory_successful(self):
        def fitness_func(bitstring):
            return 69.0

        genome_factory = PassThroughGenomeFactory(Genome.new_default_genome((5,), fitness_func))
        annealing_schedule = PassThroughAnnealingSchedule(0, True)
        optimizer_builder = BasicAnnealerOptimizerBuilder((5,),
                                                          NumpyRNG(),
                                                          genome_factory,
                                                          annealing_schedule,
                                                          fitness_func)

        optimizer = optimizer_builder \
            .add_to_initial_population_from_factory(genome_factory, 1) \
            .build()

        self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)
        self.assertIsInstance(optimizer, BasicAnnealingOptimizer)

    def test_add_to_initial_population_successful(self):
        class SpyFitnessFunc:
            def __init__(self):
                self.fitness_func_executions = 0

            def fitness_func(self, bitstring):
                self.fitness_func_executions += 1
                return 69.0

        unused_spy_fitness_function_holder = SpyFitnessFunc()
        genome_factory = PassThroughGenomeFactory(
            Genome.new_default_genome((5,), unused_spy_fitness_function_holder.fitness_func))
        annealing_schedule = PassThroughAnnealingSchedule(0, True)
        optimizer_builder = BasicAnnealerOptimizerBuilder((5,),
                                                          NumpyRNG(),
                                                          genome_factory,
                                                          annealing_schedule,
                                                          unused_spy_fitness_function_holder.fitness_func)

        spy_fitness_function_holder = SpyFitnessFunc()
        population_genome = Genome(spy_fitness_function_holder.fitness_func, (0, 0, 0, 0, 0))

        optimizer = optimizer_builder \
            .add_to_initial_population(population_genome) \
            .build()

        self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)
        self.assertEqual(spy_fitness_function_holder.fitness_func_executions, 1)
        self.assertIsInstance(optimizer, BasicAnnealingOptimizer)

    def test_build_raises_exception_with_no_population(self):
        def fitness_func(bitstring):
            self.fail()
            return 0.0

        genome_factory = PassThroughGenomeFactory(
            Genome.new_default_genome((5,), fitness_func))
        annealing_schedule = PassThroughAnnealingSchedule(0, True)
        optimizer_builder = BasicAnnealerOptimizerBuilder((5,),
                                                          NumpyRNG(),
                                                          genome_factory,
                                                          annealing_schedule,
                                                          fitness_func)

        with self.assertRaises(InitialPopulationUndefinedException):
            optimizer_builder.build()

    def test_multiple_population_add(self):
        class SpyFitnessFunc:
            def __init__(self):
                self.fitness_func_executions = 0

            def fitness_func(self, bitstring):
                self.fitness_func_executions += 1
                return 69.0

        unused_spy_fitness_function_holder = SpyFitnessFunc()
        genome_factory = PassThroughGenomeFactory(
            Genome.new_default_genome((5,), unused_spy_fitness_function_holder.fitness_func))
        annealing_schedule = PassThroughAnnealingSchedule(0, True)
        optimizer_builder = BasicAnnealerOptimizerBuilder((5,),
                                                          NumpyRNG(),
                                                          genome_factory,
                                                          annealing_schedule,
                                                          unused_spy_fitness_function_holder.fitness_func)

        spy_fitness_function_holder = SpyFitnessFunc()
        population_genome = Genome(spy_fitness_function_holder.fitness_func, (0, 0, 0, 0, 0))

        optimizer = optimizer_builder \
            .add_to_initial_population(population_genome) \
            .build()

        self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)
        self.assertEqual(spy_fitness_function_holder.fitness_func_executions, 1)
        self.assertIsInstance(optimizer, BasicAnnealingOptimizer)

        with self.assertRaises(InitialPopulationUndefinedException):
            optimizer_builder.clear().build()


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Tuple

from tunepy2.interfaces import AbstractLearner


class AbstractLearnerFactory(ABC):
    """
    The common interface that tunepy2 expects from all model builders.
    """

    @abstractmethod
    def build(self, bitstring: Tuple) -> AbstractLearner:
        """
        Builds a new model object with hyperparameters derived from the provided bitstring

        :param bitstring: a bitstring representing hyperparameters for this model
        :return: a new, untrained learner
        """
        pass



import unittest

from tunepy2 import Genome
from tunepy2.interfaces.stubs import PassThroughConvergenceCriterion, PassThroughGenomeFactory
from tunepy2.optimizers import BasicOptimizer


class TestOptimizerBasic(unittest.TestCase):
    def test_next_generation_improvement(self):
        def fitness_func_zero(bitstring):
            return 0

        def fitness_func_one(bitstring):
            return 1

        initial_candidate = Genome(fitness_func_zero, (0,))

        convergence_criterion = PassThroughConvergenceCriterion(True)
        genome_builder = PassThroughGenomeFactory(Genome(fitness_func_one, (1,)))
        optimizer = BasicOptimizer(initial_candidate, genome_builder, convergence_criterion)

        optimizer.next()

        self.assertEqual(1, optimizer.best_genome.fitness)

    def test_convergence(self):
        def fitness_func_zero(bitstring):
            return 0

        initial_candidate = Genome(fitness_func_zero, (0,))

        convergence_criterion = PassThroughConvergenceCriterion(True)
        genome_builder = PassThroughGenomeFactory(Genome(fitness_func_zero, (1,)))
        optimizer = BasicOptimizer(initial_candidate, genome_builder, convergence_criterion)

        self.assertFalse(optimizer.converged)
        optimizer.next()
        self.assertTrue(optimizer.converged)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod

from tunepy2 import Genome


class AbstractConvergenceCriterion(ABC):
    """Interface for algorithms that declare convergence of an optimizer"""

    @abstractmethod
    def converged(self, old_candidate: Genome, new_candidate: Genome) -> bool:
        """
        Determines if convergence has been reached based on current and prior Genome objects

        :param old_candidate: Genome object from prior iteration
        :param new_candidate: Genome object from current iteration
        :return: true if converged
        """
        pass


class AbstractAnnealingSchedule(AbstractConvergenceCriterion):
    """Interface for defining annealing schedules"""

    @property
    @abstractmethod
    def temperature(self) -> float:
        """ The current temperature of the annealing schedule

        :return: float value of current temperature
        """
        pass



from copy import deepcopy
from typing import Tuple

from tunepy2 import DimensionsMismatchException, TimeSeriesCrossValidatorBinException
from tunepy2.interfaces import AbstractValidator, AbstractLearner


class TimeSeriesCrossValidator(AbstractValidator):
    """
    Performs cross validation on time series data.
    """
    def __init__(self, bins: int):
        """
        Creates a new TimeSeriesCrossValidator.

        :param bins: number of bins (must be at least 2)
        """
        self._bins = bins
        self._feature_bins = []
        self._label_bins = []

    def validate(self, x, y, model: AbstractLearner) -> float:
        """
        Creates a fitness score for the provided model and data

        :param x: array-like dataset features
        :param y: array-like dataset labels
        :param model: untrained learner
        :return: a fitness score
        """
        if self._bins < 2:
            raise TimeSeriesCrossValidatorBinException

        if len(x) != len(y):
            raise DimensionsMismatchException

        self.build_bins(x, y)

        total_fitness = 0.0

        for index in range(self._bins - 1):
            model_copy = deepcopy(model)
            model_copy.fit(self._feature_bins[index], self._label_bins[index])
            model_copy.evaluate(self._feature_bins[index+1], self._label_bins[index+1])
            total_fitness += model_copy.fitness

        return total_fitness / (self._bins - 1)

    def build_bins(self, x, y):
        if len(x) != len(y):
            raise DimensionsMismatchException

        total_rows = len(y)
        extra_rows = total_rows % self._bins
        bin_size = int((total_rows - extra_rows) / self._bins)

        self._feature_bins = [None for _ in range(self._bins)]
        self._label_bins = [None for _ in range(self._bins)]

        for index in range(self._bins):
            start_slice = index * bin_size
            end_slice = (index + 1) * bin_size
            if index == self._bins - 1:
                self._feature_bins[index] = x[start_slice:]
                self._label_bins[index] = y[start_slice:]
            else:
                self._feature_bins[index] = x[start_slice:end_slice]
                self._label_bins[index] = y[start_slice:end_slice]

    @property
    def bins(self):
        return self._bins

    @property
    def feature_bins(self):
        return self._feature_bins

    @property
    def label_bins(self):
        return self._label_bins



from tunepy2 import Genome
from tunepy2.interfaces import AbstractOptimizer, AbstractGenomeFactory, AbstractConvergenceCriterion


class BasicOptimizer(AbstractOptimizer):
    """
    A very simple optimizer that builds new Genomes until convergence is satisfied.
    """
    def __init__(
            self,
            initial_candidate: Genome,
            genome_factory: AbstractGenomeFactory,
            convergence_criterion: AbstractConvergenceCriterion):
        """
        Creates a new BasicOptimizer.

        :param initial_candidate: seed Genome object
        :param genome_factory: creates new Genome objects
        :param convergence_criterion: will declare convergence once criterion is satisfied
        """
        self._candidate = initial_candidate
        self._genome_factory = genome_factory
        self._convergence_criterion = convergence_criterion
        self._converged = False

    def next(self):
        """
        Performs the next iteration of optimization.

        """
        old_candidate = self._candidate
        new_candidate = self._genome_factory.build([old_candidate])
        new_candidate.run()
        if new_candidate.fitness > old_candidate.fitness:
            self._candidate = new_candidate

        self._converged = self._convergence_criterion.converged(old_candidate, new_candidate)

    @property
    def converged(self) -> bool:
        """
        Whether or not this algorithm has converged

        :return: true when this algorithm has converged or false if not
        """
        return self._converged

    @property
    def best_genome(self) -> Genome:
        """
        The best genome so far

        :return: a Genome instance
        """
        return self._candidate



import numpy as np


def get_best_genome(genomes):
    best_fitness = float('-inf')
    best_genome = None
    for genome in genomes:
        if genome.fitness > best_fitness:
            best_genome = genome
    return best_genome


def transform_bitstring_to_layers_tuple(bitstring):
    flat_array = np.sum(bitstring, axis=0)
    return flat_array[flat_array > 0]



from typing import List, Tuple, Callable, Dict

from tunepy2.interfaces import AbstractGenomeFactory, AbstractRandomNumberGenerator
from tunepy2.internal import Genome
from copy import deepcopy


class RandomNeighborGenomeFactory(AbstractGenomeFactory):
    def __init__(
            self,
            dimensions: Tuple,
            rng: AbstractRandomNumberGenerator,
            n: int,
            fitness_func: Callable[..., float],
            *args,
            **kwargs):
        """
        Creates a new RandomNumberGenomeFactory

        :param dimensions: dimensions of bitstrings of Genome objects generated by the factory
        :param rng: random number generator
        :param n: Manhattan distance between neighbors
        :param fitness_func: fitness function passed into new Genome objects that accepts a bitstring
        :param args: will be passed into fitness_func
        :param kwargs: will be passed into fitness_func
        """
        self._n = n
        self._dimensions = dimensions
        self._rng = rng
        self._fitness_func = fitness_func
        self._args = args
        self._kwargs = kwargs

    def build(self, prior_genomes: List[Genome]) -> Genome:
        """ Builds a new Genome from a list of priors

        :param prior_genomes: list of prior Genomes
        :return: a new Genome object
        """
        def bit_flip(bit_value):
            if bit_value == 0:
                return 1
            else:
                return 0

        def recursive_bit_flip(bitstring, dimensions):
            random_index = self._rng.random_int(0, dimensions[0])
            if len(dimensions) == 1:
                bitstring[random_index] = bit_flip(bitstring[random_index])
            else:
                recursive_bit_flip(bitstring[random_index], dimensions[1:])

        selected_genome = prior_genomes[self._rng.random_int(0, len(prior_genomes))]
        candidate_genome_bitstring = deepcopy(selected_genome.bitstring)

        for _ in range(self._n):
            recursive_bit_flip(candidate_genome_bitstring, self._dimensions)

        return Genome(self._fitness_func, candidate_genome_bitstring, *self._args, **self._kwargs)

    @property
    def dimensions(self) -> Tuple:
        """

        :return: the dimensions of Genome objects returned by this factory
        """
        return self._dimensions



from tunepy2.interfaces import AbstractGenomeComparer


class PassThroughComparer(AbstractGenomeComparer):
    """
    Returns the first item passed into it.
    """

    def compare(self, genomes):
        """
        Compares a set of models naively.
        :param genomes: Array-like of genomes.
        :return: THe first model at the zero index of models.
        """
        return genomes[0]


class IncrementalComparer(AbstractGenomeComparer):
    """
    Returns the next item from the list.
    """

    def __init__(self):
        self._index = 0

    def compare(self, genomes):
        next_index = self._index % len(genomes)
        self._index += 1
        return genomes[next_index]



from typing import Tuple, Callable

from tunepy2 import InitialPopulationUndefinedException, Genome
from tunepy2.interfaces import AbstractOptimizerBuilder, AbstractRandomNumberGenerator, AbstractGenomeFactory, \
    AbstractConvergenceCriterion, AbstractOptimizer
from tunepy2.optimizers import BasicGeneticOptimizer


class BasicGeneticOptimizerBuilder(AbstractOptimizerBuilder):
    """
    Builds new instances of BasicGeneticOptimizer.
    """

    def __init__(
            self,
            dimensions: Tuple,
            rng: AbstractRandomNumberGenerator,
            new_candidate_genome_factory: AbstractGenomeFactory,
            convergence_criterion: AbstractConvergenceCriterion,
            fitness_func: Callable[..., float],
            *args,
            **kwargs):
        """
        Creates a new BasicGeneticOptimizerBuilder.

        :param dimensions: dimensions of bitstrings
        :param rng: a random number generator
        :param new_candidate_genome_factory: factory for building new Genome objects
        :param convergence_criterion: criterion to use for declaring convergence
        :param fitness_func: fitness function passed into new Genome objects
        :param args: will be passed into fitness_func
        :param kwargs: will be passed into fitness_func
        """
        self._dimensions = dimensions
        self._rng = rng
        self._fitness_func = fitness_func
        self._new_candidate_genome_factory = new_candidate_genome_factory
        self._convergence_criterion = convergence_criterion
        self._args = args
        self._kwargs = kwargs
        self._population = []

    def add_to_initial_population_from_factory(
            self,
            genome_factory: AbstractGenomeFactory,
            n: int) -> AbstractOptimizerBuilder:
        """
        Adds a given number of Genomes to the initial population of this optimizer from a factory

        :param genome_factory: the factory that will generate the Genome
        :param n: number of instances to add
        :return: self
        """
        base_genome = Genome.new_default_genome(genome_factory.dimensions,
                                                self._fitness_func,
                                                *self._args,
                                                **self._kwargs)
        self._population += [genome_factory.build([base_genome]) for _ in range(n)]
        return self

    def add_to_initial_population(self, genome: Genome) -> AbstractOptimizerBuilder:
        """
        Adds a single Genome instance to the initial population of this optimizer

        :param genome: instance to add
        :return: self
        """
        self._population.append(genome)
        return self

    def build(self) -> AbstractOptimizer:
        """
        Constructs a new optimizer

        :return: a new optimizer
        """
        for genome in self._population:
            genome.run()

        if len(self._population) < 1:
            raise InitialPopulationUndefinedException

        new_optimizer = BasicGeneticOptimizer(self._population,
                                              self._new_candidate_genome_factory,
                                              self._convergence_criterion)

        return new_optimizer

    def clear(self) -> AbstractOptimizerBuilder:
        """
        Clears the initial population of this builder

        :return: self
        """
        self._population = []
        return self



from typing import Tuple

from tunepy2.interfaces import AbstractValidator, AbstractLearner
from tunepy2.internal import *
import numpy as np
from copy import deepcopy


class CrossValidator(AbstractValidator):
    """
    Trains a set number of models on subsets of the provided data and returns the average fitness
    of the trained models.
    """

    def __init__(self, bins: int):
        """
        Creates a new CrossValidator.

        :param bins: number of bins (must be at least 2)
        """
        self._bins = bins
        self._feature_bins_train = []
        self._feature_bins_test = []
        self._label_bins_train = []
        self._label_bins_test = []

    def validate(self, x, y, model: AbstractLearner) -> float:
        """
        Creates a fitness score for the provided model and data

        :param x: array-like dataset features
        :param y: array-like dataset labels
        :param model: untrained learner
        :return: a fitness score
        """
        if self._bins < 2:
            raise CrossValidatorBinException

        if len(x) != len(y):
            raise DimensionsMismatchException

        self.build_test_bins(x, y)
        self.build_train_bins(x, y)

        total_fitness = 0.0

        for index in range(self._bins):
            model_copy = deepcopy(model)
            model_copy.fit(self._feature_bins_train[index], self._label_bins_train[index])
            model_copy.evaluate(self._label_bins_test[index], self._label_bins_train[index])
            total_fitness += model_copy.fitness
        return float(total_fitness / self._bins)

    def build_test_bins(self, x, y):
        if len(x) != len(y):
            raise DimensionsMismatchException

        total_rows = len(y)
        extra_rows = total_rows % self._bins
        bin_size = int((total_rows - extra_rows) / self._bins)

        self._feature_bins_test = [None for _ in range(self._bins)]
        self._label_bins_test = [None for _ in range(self._bins)]

        for index in range(self._bins):
            start_slice = index * bin_size
            end_slice = (index + 1) * bin_size
            if index == self._bins - 1:
                self._feature_bins_test[index] = x[start_slice:]
                self._label_bins_test[index] = y[start_slice:]
            else:
                self._feature_bins_test[index] = x[start_slice:end_slice]
                self._label_bins_test[index] = y[start_slice:end_slice]

    def build_train_bins(self, x, y):
        if len(x) != len(y):
            raise DimensionsMismatchException

        total_rows = len(y)
        extra_rows = total_rows % self._bins
        bin_size = int((total_rows - extra_rows) / self._bins)

        self._feature_bins_train = [None for _ in range(self._bins)]
        self._label_bins_train = [None for _ in range(self._bins)]

        for index in range(self._bins):
            start_slice = index * bin_size
            end_slice = (index + 1) * bin_size
            if index == 0:
                self._feature_bins_train[index] = x[end_slice:]
                self._label_bins_train[index] = y[end_slice:]
            elif index == self._bins - 1:
                self._feature_bins_train[index] = x[:start_slice]
                self._label_bins_train[index] = y[:start_slice]
            else:
                self._feature_bins_train[index] = np.concatenate((x[:start_slice], x[end_slice:]), axis=0)
                self._label_bins_train[index] = np.concatenate((y[:start_slice], y[end_slice:]), axis=0)

    @property
    def bins(self):
        return self._bins

    @property
    def training_data_features(self):
        return self._feature_bins_train

    @property
    def training_data_labels(self):
        return self._label_bins_train

    @property
    def test_data_features(self):
        return self._feature_bins_test

    @property
    def test_data_labels(self):
        return self._label_bins_test






from tunepy2.interfaces import AbstractLearner


class ConstantFitnessLearner(AbstractLearner):
    """
    A learner with a fitness defined on construction.
    """

    def __init__(self, fitness):
        self._fitness = fitness

    def fit(self, x_train, y_train):
        pass

    def evaluate(self, x_test, y_test):
        return self.fitness

    @property
    def fitness(self):
        return self._fitness



from __future__ import annotations

from abc import ABC, abstractmethod
from tunepy2 import Genome
from tunepy2.interfaces import AbstractGenomeFactory, AbstractOptimizer


class AbstractOptimizerBuilder(ABC):
    """
    Common interface for all optimizers.
    """

    @abstractmethod
    def add_to_initial_population_from_factory(
            self,
            genome_factory: AbstractGenomeFactory,
            n: int) -> AbstractOptimizerBuilder:
        """
        Adds a given number of Genomes to the initial population of this optimizer from a factory

        :param genome_factory: the factory that will generate the Genome
        :param n: number of instances to add
        :return: self
        """
        pass

    @abstractmethod
    def add_to_initial_population(self, genome: Genome) -> AbstractOptimizerBuilder:
        """
        Adds a single Genome instance to the initial population of this optimizer

        :param genome: instance to add
        :return: self
        """
        pass

    @abstractmethod
    def build(self) -> AbstractOptimizer:
        """
        Constructs a new optimizer

        :return: a new optimizer
        """
        pass

    @abstractmethod
    def clear(self) -> AbstractOptimizerBuilder:
        """
        Clears the initial population of this builder

        :return: self
        """
        pass



from typing import List

from tunepy2.interfaces import AbstractOptimizer, AbstractGenomeFactory, AbstractConvergenceCriterion
from tunepy2 import get_best_genome, Genome


class BasicGeneticOptimizer(AbstractOptimizer):
    """
    Optimizes fitness using successive refinements of a random population of solutions.
    """

    def __init__(
            self,
            initial_population: List[Genome],
            genome_factory: AbstractGenomeFactory,
            convergence_criterion: AbstractConvergenceCriterion):
        """
        Creates a new BasicGeneticOptimizer.

        :param initial_population: initial seed of Genome objects
        :param genome_factory: creates new Genome objects
        :param convergence_criterion: will declare convergence once criterion is satisfied
        """
        self._population = initial_population
        self._genome_factory = genome_factory
        self._convergence_criterion = convergence_criterion
        self._max_fitness = float('-inf')
        self._best_genome = None
        self._converged = False

    def next(self):
        """
        Performs the next iteration of optimization.

        """
        old_population = self._population

        self._population = []

        for index in range(len(old_population)):
            new_genome = self._genome_factory.build(old_population)
            self._population.append(new_genome)
            new_genome.run()
            if new_genome.fitness > self._max_fitness:
                self._best_genome = new_genome
                self._max_fitness = self._best_genome.fitness

        self._converged = self._convergence_criterion.converged(get_best_genome(old_population), self.best_genome)

    @property
    def converged(self) -> bool:
        """
        Whether or not this algorithm has converged

        :return: true when this algorithm has converged or false if not
        """
        return self._converged

    @property
    def best_genome(self) -> Genome:
        """
        The best genome so far

        :return: a Genome instance
        """
        if self._best_genome is None:
            self._best_genome = get_best_genome(self._population)
        return self._best_genome



import unittest
from tunepy2.validators import CrossValidator
from tunepy2 import *
from tunepy2.interfaces.stubs import *


class TestCrossValidator(unittest.TestCase):
    def test_bin_creation_error(self):
        validator = CrossValidator(1)

        with self.assertRaises(CrossValidatorBinException):
            validator.validate(None, None, None)

    def test_test_bins_error(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = CrossValidator(5)

        with self.assertRaises(DimensionsMismatchException):
            validator.build_test_bins(data_features, data_labels)

    def test_train_bins_error(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = CrossValidator(5)

        with self.assertRaises(DimensionsMismatchException):
            validator.build_train_bins(data_features, data_labels)

    def test_data_mismatch_error(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = CrossValidator(5)

        with self.assertRaises(DimensionsMismatchException):
            validator.validate(data_features, data_labels, None)

    def test_evaluation_bin_sizes(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4, 5]

        validator = CrossValidator(5)
        validator.build_test_bins(data_features, data_labels)

        self.assertEqual(5, len(validator.test_data_features))
        self.assertEqual(5, len(validator.test_data_labels))

        self.assertEqual(1, len(validator.test_data_features[0]))
        self.assertEqual(1, len(validator.test_data_labels[0]))

        self.assertEqual(2, len(validator.test_data_features[-1]))
        self.assertEqual(2, len(validator.test_data_labels[-1]))

    def test_train_bin_sizes(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4, 5]

        validator = CrossValidator(5)
        validator.build_train_bins(data_features, data_labels)

        self.assertEqual(5, len(validator.training_data_features))
        self.assertEqual(5, len(validator.training_data_labels))

        self.assertEqual(5, len(validator.training_data_features[0]))
        self.assertEqual(5, len(validator.training_data_labels[0]))

        self.assertEqual(4, len(validator.training_data_features[-1]))
        self.assertEqual(4, len(validator.training_data_labels[-1]))

    def test_validation_result(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = CrossValidator(5)
        fitness = validator.validate(data_features, data_labels, ConstantFitnessLearner(69.0))
        self.assertAlmostEqual(69.0, fitness)


if __name__ == '__main__':
    unittest.main()



class CrossValidatorBinException(Exception):
    pass


class TimeSeriesCrossValidatorBinException(Exception):
    pass


class DimensionsMismatchException(Exception):
    pass


class TrainTestSplitException(Exception):
    pass


class InitialPopulationUndefinedException(Exception):
    pass



from tunepy2.interfaces import AbstractConvergenceCriterion, AbstractAnnealingSchedule


class PassThroughConvergenceCriterion(AbstractConvergenceCriterion):
    """
    Returns the value passed in during instantiation.
    """

    def converged(self, old_candidate, new_candidate):
        return self._return_value

    def __init__(self, return_value):
        self._return_value = return_value


class PassThroughAnnealingSchedule(AbstractAnnealingSchedule):
    @property
    def temperature(self):
        return self._temperature

    def converged(self, old_candidate, new_candidate):
        return self._converged_value

    def __init__(self, temperature_value, converged_value):
        self._temperature = temperature_value
        self._converged_value = converged_value



import unittest

from tunepy2 import Genome
from tunepy2.comparison import RouletteWheelComparer
from tunepy2.interfaces import AbstractRandomNumberGenerator


class TestRouletteWheelComparer(unittest.TestCase):

    def test_returns_best_model(self):
        class CustomRandom(AbstractRandomNumberGenerator):

            def random_int_array(self, minimum, maximum, shape):
                return [0]

            def random(self):
                return 0.0

        rng = CustomRandom()
        comparer = RouletteWheelComparer(rng)
        genomes = [Genome.new_default_genome((10-index,), lambda bitstring: len(bitstring)) for index in range(10)]

        for genome in genomes:
            genome.run()

        returned_model = comparer.compare(genomes)
        self.assertEqual(10, returned_model.fitness)

    def test_returns_worst_model(self):
        class CustomRandom(AbstractRandomNumberGenerator):

            def random_int_array(self, minimum, maximum, shape):
                return [9]

            def random(self):
                return 0.0

        rng = CustomRandom()
        comparer = RouletteWheelComparer(rng)
        genomes = [Genome.new_default_genome((10-index,), lambda bitstring: len(bitstring)) for index in range(10)]

        for genome in genomes:
            genome.run()

        returned_model = comparer.compare(genomes)
        self.assertEqual(1, returned_model.fitness)

    def test_returns_median_model(self):
        class CustomRandom(AbstractRandomNumberGenerator):

            def random_int_array(self, minimum, maximum, shape):
                return [5]

            def random(self):
                return 0.0

        rng = CustomRandom()
        comparer = RouletteWheelComparer(rng)
        genomes = [Genome.new_default_genome((10-index,), lambda bitstring: len(bitstring)) for index in range(10)]

        for genome in genomes:
            genome.run()

        returned_model = comparer.compare(genomes)
        self.assertEqual(5, returned_model.fitness)


if __name__ == '__main__':
    unittest.main()



from tunepy2.interfaces.stubs.stub_rngs import *
from tunepy2.interfaces.stubs.stub_learners import *
from tunepy2.interfaces.stubs.stub_model_comparers import *
from tunepy2.interfaces.stubs.stub_optimizers import *
from tunepy2.interfaces.stubs.stub_model_factories import *
from tunepy2.interfaces.stubs.stub_validators import *
from tunepy2.interfaces.stubs.stub_convergence_criteria import *
from tunepy2.interfaces.stubs.stub_genome_factories import *



from tunepy2.interfaces import AbstractRandomNumberGenerator
import numpy as np


class NumpyOnesRandom(AbstractRandomNumberGenerator):
    """
    Returns a numpy array of ones in whatever shape is provided.
    """

    def random(self):
        return 1.0

    def random_int_array(self, minimum, maximum, shape):
        return np.ones(shape=shape, dtype='int')


class NumpyPassThroughRandom(AbstractRandomNumberGenerator):
    """
    Returns an input passed on instantiation.
    """

    def __init__(self, constant_random):
        self._constant_random = constant_random

    def random_int_array(self, minimum, maximum, shape):
        return self._constant_random * np.ones(shape=shape, dtype='int')

    def random(self):
        return self._constant_random


class NumpyCustomRandom(AbstractRandomNumberGenerator):
    """
    Returns constructor defined values for all methods.
    """

    def __init__(self, random_float, random_integer):
        self._float = random_float
        self._int = random_integer

    def random_int_array(self, minimum, maximum, shape):
        return self._int * np.ones(shape=shape, dtype='int')

    def random(self):
        return self._float



from abc import ABC, abstractmethod
from typing import Tuple


class AbstractRandomNumberGenerator(ABC):
    """
    The common interface for that tunepy2 expects from all random number generators.
    """

    @abstractmethod
    def random_int_array(self, minimum: int, maximum: int, shape: Tuple) -> Tuple:
        """
        Builds an array-like structure of random integers

        :param minimum: minimum value (inclusive)
        :param maximum: maximum value (exclusive)
        :param shape: the shape of the output
        :return: a collection of integers
        """
        pass

    @abstractmethod
    def random(self) -> float:
        """
        Returns a random number between 0 and 1

        :return: random number between 0 and 1
        """
        pass

    def random_int(self, minimum: int, maximum: int) -> int:
        """
        Returns a single random integer

        :param minimum: minimum value (inclusive)
        :param maximum: maximum value (exclusive)
        :return: random integer
        """
        return self.random_int_array(minimum, maximum, shape=(1,))[0]



import unittest
from tunepy2 import new_random_restart_hill_climber, new_simulated_annealer, new_hill_climber, new_genetic_optimizer
from tunepy2.optimizers.meta import BasicRestartOptimizer
from tunepy2.optimizers import BasicAnnealingOptimizer, BasicOptimizer, BasicGeneticOptimizer


class TestPrebuiltOptimizers(unittest.TestCase):
    def test_creates_random_restart_hill_climber(self):
        def fitness_func(bitstring):
            return 69.0

        optimizer = new_random_restart_hill_climber(
            (5,),
            7,
            10,
            1e-5,
            fitness_func)

        self.assertIsInstance(optimizer, BasicRestartOptimizer)
        self.assertFalse(optimizer.converged)

        for _ in range(6):
            optimizer.next()
            self.assertFalse(optimizer.converged)
            self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)

        optimizer.next()
        self.assertTrue(optimizer.converged)

    def test_creates_simulated_annealer(self):
        def fitness_func(bitstring):
            return 69.0

        optimizer = new_simulated_annealer(
            (5,),
            1,
            10,
            3,
            0.5,
            fitness_func)

        self.assertIsInstance(optimizer, BasicAnnealingOptimizer)
        self.assertFalse(optimizer.converged)

        for _ in range(2):
            optimizer.next()
            self.assertFalse(optimizer.converged)
            self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)

        optimizer.next()
        self.assertTrue(optimizer.converged)

    def test_creates_hill_climber(self):
        def fitness_func(bitstring):
            return 69.0

        optimizer = new_hill_climber(
            (5,),
            10,
            1e-5,
            fitness_func)

        self.assertIsInstance(optimizer, BasicOptimizer)
        self.assertFalse(optimizer.converged)

        for _ in range(9):
            optimizer.next()
            self.assertFalse(optimizer.converged)
            self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)

        optimizer.next()
        self.assertTrue(optimizer.converged)

    def test_creates_genetic_optimizer(self):
        def fitness_func(bitstring):
            return 69.0

        optimizer = new_genetic_optimizer(
            (5,),
            100,
            0.2,
            10,
            1e-5,
            fitness_func)

        self.assertIsInstance(optimizer, BasicGeneticOptimizer)
        self.assertFalse(optimizer.converged)

        for _ in range(9):
            optimizer.next()
            self.assertFalse(optimizer.converged)
            self.assertAlmostEqual(optimizer.best_genome.fitness, 69.0)

        optimizer.next()
        self.assertTrue(optimizer.converged)


if __name__ == '__main__':
    unittest.main()



from typing import Tuple

from tunepy2.interfaces import AbstractLearner
from sklearn.neural_network import MLPRegressor


class SklearnMlpRegressorLearner(AbstractLearner):
    """
    Wraps Sklearn MLPRegressor for hyperparameter search.
    """
    def __init__(
            self,
            *args,
            **kwargs):
        """
        Creates a new SklearnMlpRegressorLearner.

        :param args: will be passed into MLPRegressor during construction
        :param kwargs: will be passed into MLPRegressor during construction
        """
        self._learner = MLPRegressor(*args, **kwargs)
        self._fitness = float("-inf")

    def fit(self, x_train, y_train):
        """
        Trains the model on a labeled dataset

        :param x_train: table of features
        :param y_train: vector of labels
        """
        self._learner.fit(x_train, y_train)

    def evaluate(self, x_test: Tuple, y_test: Tuple) -> float:
        """
        Returns a fitness score after predicting labels on a dataset

        :param x_test: table of features.
        :param y_test: vector of labels for fitness evaluation.
        :return: A fitness score
        """
        self._fitness = self._learner.score(x_test, y_test)
        return self._fitness

    @property
    def fitness(self) -> float:
        """
        The fitness score of this machine learning model

        :return: The fitness score of this machine learning model
        """
        return self._fitness



from typing import Tuple, Callable, List, Dict

from tunepy2 import Genome
from tunepy2.interfaces import AbstractGenomeFactory, AbstractRandomNumberGenerator


class RandomGenomeFactory(AbstractGenomeFactory):
    def __init__(
            self,
            dimensions: Tuple,
            rng: AbstractRandomNumberGenerator,
            fitness_func: Callable[..., float],
            *args,
            **kwargs):
        """
        Creates a new RandomGenomeFactory

        :param dimensions: dimensions of bitstrings of Genome objects generated by the factory
        :param rng: random number generator
        :param fitness_func: fitness function passed into new Genome objects that accepts a bitstring
        :param args: will be passed into fitness_func
        :param kwargs: will be passed into fitness_func
        """
        self._rng = rng
        self._dimensions = dimensions
        self._fitness_func = fitness_func
        self._args = args
        self._kwargs = kwargs

    def build(self, prior_genomes: List[Genome]) -> Genome:
        """ Builds a new Genome from a list of priors

        :param prior_genomes: list of prior Genomes
        :return: a new Genome object
        """
        bitstring = self._rng.random_int_array(0, 1, self._dimensions)
        return Genome(self._fitness_func, bitstring, *self._args, **self._kwargs)

    @property
    def dimensions(self) -> Tuple:
        """

        :return: the dimensions of Genome objects returned by this factory
        """
        return self._dimensions



from tunepy2.interfaces import AbstractOptimizer


class ConvergedOptimizer(AbstractOptimizer):
    """
    Always appears to have converged.
    """

    @property
    def best_genome(self):
        return None

    @property
    def converged(self):
        return True

    def next(self):
        pass



from tunepy2.interfaces import AbstractValidator, AbstractLearner
from tunepy2.internal import DimensionsMismatchException, TrainTestSplitException


class TrainTestValidator(AbstractValidator):
    """
    Trains models on a portion of the data and evaluates performs on the other.
    """
    def __init__(self, train_test_ratio: float):
        """
        Creates a new TrainTestValidator.

        :param train_test_ratio: portion of data used for training (must be on interval (0,1))
        """
        self._train_test_ratio = train_test_ratio
        self._x = []
        self._y = []
        self._test_features = []
        self._test_labels = []
        self._train_features = []
        self._train_labels = []

    def validate(self, x, y, model: AbstractLearner) -> float:
        """
        Creates a fitness score for the provided model and data

        :param x: array-like dataset features
        :param y: array-like dataset labels
        :param model: untrained learner
        :return: a fitness score
        """
        if len(x) != len(y):
            raise DimensionsMismatchException

        self.split_data(x, y)
        model.fit(self.training_data_features, self.training_data_labels)
        return float(model.evaluate(self.test_data_features, self.training_data_labels))

    def split_data(self, x, y):
        if len(x) != len(y):
            raise DimensionsMismatchException

        self._x = x
        self._y = y

        size_train = int(self._train_test_ratio * len(x))

        if size_train <= 0 or size_train >= len(x):
            raise TrainTestSplitException

        self._train_features = self._x[:size_train]
        self._train_labels = self._y[:size_train]
        self._test_features = self._x[size_train:]
        self._test_labels = self._y[size_train:]

    @property
    def test_data_features(self):
        return self._test_features

    @property
    def test_data_labels(self):
        return self._test_labels

    @property
    def training_data_features(self):
        return self._train_features

    @property
    def training_data_labels(self):
        return self._train_labels



from typing import Tuple

from tunepy2.interfaces import AbstractRandomNumberGenerator
import numpy as np


class NumpyRNG(AbstractRandomNumberGenerator):
    """
    A tunepy2 wrapper for the number random number generator.
    """

    def __init__(
            self,
            seed=None,
            integer_type='int32'):
        """
        Creates a new numpy-based random number generator.

        :param seed: seed that will be passed into numpy (global to numpy)
        :param integer_type: numpy type of integers generated by this random number generator
        """
        if seed is not None:
            np.random.seed(seed)
        self._integer_type = integer_type

    def random(self) -> float:
        """
        Returns a random number between 0 and 1

        :return: random number between 0 and 1
        """
        return np.random.random()

    def random_int_array(self, minimum: int, maximum: int, shape: Tuple) -> Tuple:
        """
        Builds an array-like structure of random integers

        :param minimum: minimum value (inclusive)
        :param maximum: maximum value (exclusive)
        :param shape: the shape of the output
        :return: a collection of integers
        """
        return np.random.randint(minimum, maximum, size=shape, dtype=self._integer_type)



from abc import ABC, abstractmethod

from tunepy2 import Genome


class AbstractOptimizer(ABC):
    """
    The common interface for that tunepy2 expects from all optimizing algorithms.
    """

    @abstractmethod
    def next(self):
        """
        Performs the next iteration of optimization.

        """
        pass

    @property
    @abstractmethod
    def converged(self) -> bool:
        """
        Whether or not this algorithm has converged

        :return: true when this algorithm has converged or false if not
        """
        pass

    @property
    @abstractmethod
    def best_genome(self) -> Genome:
        """
        The best genome so far

        :return: a Genome instance
        """
        pass



from tunepy2 import Genome
from tunepy2.interfaces import AbstractAnnealingSchedule


class ExponentialAnnealingSchedule(AbstractAnnealingSchedule):
    def __init__(self, initial_temperature: float, minimum_temperature: float, degradation_multiplier: float):
        """ Creates a new ExponentialAnnealingSchedule

        :param initial_temperature: starting temperature
        :param minimum_temperature: minimum possible temperature
        :param degradation_multiplier: temperature multiplier (must be less than 1.0)
        """
        self._temperature = initial_temperature
        self._minimum_temperature = minimum_temperature
        self._degradation_multiplier = degradation_multiplier

    @property
    def temperature(self) -> float:
        """ The current temperature of the annealing schedule

        :return: float value of current temperature
        """
        return self._temperature

    def converged(self, old_candidate: Genome, new_candidate: Genome) -> bool:
        """ Determines if convergence has been reached based on current and prior Genome objects

        :param old_candidate: Genome object from prior iteration
        :param new_candidate: Genome object from current iteration
        :return: true if converged
        """
        return_value = self._temperature < self._minimum_temperature
        self._temperature *= self._degradation_multiplier
        return return_value



import unittest

from tunepy2.interfaces.stubs import ConstantFitnessLearner
from tunepy2.validators import TimeSeriesCrossValidator
from tunepy2 import TimeSeriesCrossValidatorBinException, DimensionsMismatchException


class TestTimeSeriesCrossValidator(unittest.TestCase):
    def test_bin_creation_error(self):
        validator = TimeSeriesCrossValidator(1)

        with self.assertRaises(TimeSeriesCrossValidatorBinException):
            validator.validate(None, None, None)

    def test_dimensions_mismatch_error_on_build(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TimeSeriesCrossValidator(5)

        with self.assertRaises(DimensionsMismatchException):
            validator.build_bins(data_features, data_labels)

    def test_dimensions_mismatch_error_on_validate(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TimeSeriesCrossValidator(5)

        with self.assertRaises(DimensionsMismatchException):
            validator.validate(data_features, data_labels, None)

    def test_bin_sizes(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4, 5]

        validator = TimeSeriesCrossValidator(5)
        validator.build_bins(data_features, data_labels)

        self.assertEqual(5, len(validator.feature_bins))
        self.assertEqual(5, len(validator.label_bins))

        self.assertEqual(1, len(validator.feature_bins[0]))
        self.assertEqual(1, len(validator.label_bins[0]))

        self.assertEqual(2, len(validator.feature_bins[-1]))
        self.assertEqual(2, len(validator.label_bins[-1]))

    def test_validation_result(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TimeSeriesCrossValidator(5)
        fitness = validator.validate(data_features, data_labels, ConstantFitnessLearner(69.0))
        self.assertAlmostEqual(69.0, fitness)

    def test_bin_count(self):
        validator = TimeSeriesCrossValidator(10)
        self.assertEqual(10, validator.bins)

if __name__ == '__main__':
    unittest.main()



from tunepy2 import Genome
from tunepy2.interfaces import AbstractOptimizer, AbstractGenomeFactory, AbstractAnnealingSchedule, \
    AbstractRandomNumberGenerator
from math import exp


class BasicAnnealingOptimizer(AbstractOptimizer):
    """
    Performs a randomized optimization of by simulated annealing.
    """

    def __init__(
            self,
            initial_candidate: Genome,
            genome_factory: AbstractGenomeFactory,
            annealing_schedule: AbstractAnnealingSchedule,
            rng: AbstractRandomNumberGenerator):
        """
        Creates a new BasicAnnealingOptimizer.

        :param initial_candidate: seed Genome object
        :param genome_factory: builds new Genome objects
        :param annealing_schedule: temperature schedule for simulated annealing
        :param rng: a random number generator
        """
        self._candidate = initial_candidate
        self._genome_factory = genome_factory
        self._annealing_schedule = annealing_schedule
        self._rng = rng
        self._max_fitness = float('-inf')
        self._best_genome = initial_candidate
        self._converged = False

    def next(self):
        """
        Performs the next iteration of optimization.

        """
        old_candidate = self._candidate
        new_candidate = self._genome_factory.build([old_candidate])
        new_candidate.run()

        if new_candidate.fitness > old_candidate.fitness:
            self._candidate = new_candidate
        else:
            acceptance_probability = exp((new_candidate.fitness - old_candidate.fitness) / self.temperature)
            if self._rng.random() < acceptance_probability:
                self._candidate = new_candidate

        if self._candidate.fitness > self._max_fitness:
            self._best_genome = self._candidate
            self._max_fitness = self._candidate.fitness

        self._converged = self._annealing_schedule.converged(old_candidate, self._candidate)

    @property
    def converged(self) -> bool:
        """
        Whether or not this algorithm has converged

        :return: true when this algorithm has converged or false if not
        """
        return self._converged

    @property
    def best_genome(self) -> Genome:
        """
        The best genome so far

        :return: a Genome instance
        """
        return self._best_genome

    @property
    def temperature(self) -> float:
        """
        The current temperature of the BasicAnnealingOptimizer

        :return: the current temperature
        """
        return self._annealing_schedule.temperature



import unittest
from tunepy2.convergence import ConsecutiveNonImprovement


class TestConsecutiveNonImprovement(unittest.TestCase):
    def test_divergence(self):
        class FakeGenome:
            def __init__(self, fitness):
                self._fitness = fitness

            @property
            def fitness(self):
                return self._fitness

        convergence = ConsecutiveNonImprovement(5, 2)

        old_genome = FakeGenome(-1)
        for index in range(20):
            new_genome = FakeGenome(index)
            self.assertFalse(convergence.converged(old_genome, new_genome))
            old_genome = new_genome

    def test_convergence(self):
        class FakeGenome:
            def __init__(self, fitness):
                self._fitness = fitness

            @property
            def fitness(self):
                return self._fitness

        convergence = ConsecutiveNonImprovement(5, 1e-5)

        old_genome = FakeGenome(-1)
        for index in range(4):
            self.assertFalse(convergence.converged(old_genome, old_genome))
        self.assertTrue(convergence.converged(old_genome, old_genome))


if __name__ == '__main__':
    unittest.main()



from typing import List, Dict, Tuple

from tunepy2 import transform_bitstring_to_layers_tuple
from tunepy2.interfaces import AbstractLearnerFactory, AbstractRandomNumberGenerator, AbstractLearner
from tunepy2.learners import SklearnMlpClassifierLearner


class SklearnMlpClassifierFactory(AbstractLearnerFactory):
    """
    Transforms bitstrings into SklearnMlpClassifierLearner.
    """
    def __init__(
            self,
            *args,
            **kwargs):
        """
        Creates a new SklearnMlpClassifierFactory.

        :param args: will be passed into the MLPClassifier
        :param kwargs: will be passed into the MLPClassifier
        """
        self._args = args
        self._kwargs = kwargs

    def build(self, bitstring: Tuple) -> AbstractLearner:
        """
        Builds a new model object with hyperparameters derived from the provided bitstring

        :param bitstring: a bitstring representing hyperparameters for this model
        :return: a new, untrained learner
        """
        layers = transform_bitstring_to_layers_tuple(bitstring)
        return SklearnMlpClassifierLearner(layers=layers, *self._args, **self._kwargs)



from abc import ABC, abstractmethod
from typing import List, Tuple

from tunepy2 import Genome


class AbstractGenomeFactory(ABC):
    """ Interface for constructing new Genome objects based on prior Genomes """

    @abstractmethod
    def build(self, prior_genomes: List[Genome]) -> Genome:
        """ Builds a new Genome from a list of priors

        :param prior_genomes: list of prior Genomes
        :return: a new Genome object
        """
        pass

    @property
    @abstractmethod
    def dimensions(self) -> Tuple:
        """

        :return: the dimensions of Genome objects returned by this factory
        """
        pass




from tunepy2.optimizers.genetic_optimizer_basic import BasicGeneticOptimizer
from tunepy2.optimizers.annealer_optimizer_basic import BasicAnnealingOptimizer
from tunepy2.optimizers.optimizer_basic import BasicOptimizer



from tunepy2.genome_factory.ga_genome_factory import UniformCrossoverGenomeFactory, SinglePointCrossoverGenomeFactory
from tunepy2.genome_factory.neighbor_genome_factory import RandomNeighborGenomeFactory
from tunepy2.genome_factory.random_genome_factory import RandomGenomeFactory



import unittest
from tunepy2.interfaces.stubs import IncrementalComparer, NumpyOnesRandom, NumpyPassThroughRandom
from tunepy2.internal import DimensionsMismatchException, Genome
from tunepy2.genome_factory import UniformCrossoverGenomeFactory
import numpy as np


class TestUniformCrossoverGenomeBuilder(unittest.TestCase):
    def test_incorrect_dimensions_passed_in(self):
        def fitness_func(bitstring):
            return 0.0

        left = Genome(fitness_func, (0, 0, 0, 0, 0))
        right = Genome(fitness_func, (1, 1, 1, 1, 1))
        dimensions = (4,)

        comparer = IncrementalComparer()
        rng = NumpyOnesRandom()
        genome_builder = UniformCrossoverGenomeFactory(dimensions, rng, 0.0, comparer, fitness_func)

        with self.assertRaises(DimensionsMismatchException):
            genome_builder.build([left, right])

    def test_mutation(self):
        def fitness_func(bitstring):
            return 0.0

        left = Genome(fitness_func, (0, 0, 0, 0, 0))
        right = Genome(fitness_func, (0, 0, 0, 0, 0))
        dimensions = (5,)

        comparer = IncrementalComparer()
        rng = NumpyOnesRandom()
        genome_builder = UniformCrossoverGenomeFactory(dimensions, rng, 1.0, comparer, fitness_func)

        new_genome = genome_builder.build([left, right])

        self.assertEqual([1, 1, 1, 1, 1], list(new_genome.bitstring))

    def test_left_inheritance(self):
        def fitness_func(bitstring):
            return 0.0

        left = Genome(fitness_func, 2 * np.ones((5, 5)))
        right = Genome(fitness_func, 3 * np.ones((5, 5)))
        dimensions = (5, 5)

        comparer = IncrementalComparer()
        rng = NumpyPassThroughRandom(0.2)
        genome_builder = UniformCrossoverGenomeFactory(dimensions, rng, 0.0, comparer, fitness_func)

        new_genome = genome_builder.build([left, right])

        self.assertTrue(np.array_equal(left.bitstring, new_genome.bitstring))

    def test_right_inheritance(self):
        def fitness_func(bitstring):
            return 0.0

        left = Genome(fitness_func, 2 * np.ones((5, 5)))
        right = Genome(fitness_func, 3 * np.ones((5, 5)))
        dimensions = (5, 5)

        comparer = IncrementalComparer()
        rng = NumpyPassThroughRandom(0.8)
        genome_builder = UniformCrossoverGenomeFactory(dimensions, rng, 0.0, comparer, fitness_func)

        new_genome = genome_builder.build([left, right])

        self.assertTrue(np.array_equal(right.bitstring, new_genome.bitstring))


if __name__ == '__main__':
    unittest.main()



from typing import Tuple, Callable

from tunepy2 import Genome
from tunepy2.interfaces import AbstractOptimizer
from tunepy2.optimizers.meta import BasicRestartOptimizer
from tunepy2.optimizers import BasicAnnealingOptimizer, BasicOptimizer, BasicGeneticOptimizer
from tunepy2.convergence import Iterations, ConsecutiveNonImprovement, ExponentialAnnealingSchedule
from tunepy2.optimizers.builders import BasicOptimizerBuilder
from tunepy2.genome_factory import RandomGenomeFactory, \
    RandomNeighborGenomeFactory, \
    SinglePointCrossoverGenomeFactory
from tunepy2.random import NumpyRNG
from tunepy2.comparison import RouletteWheelComparer


def new_random_restart_hill_climber(
        dimensions: Tuple,
        restarts: int,
        convergence_iterations: int,
        epsilon: float,
        fitness_func: Callable[..., float],
        *args,
        **kwargs) -> AbstractOptimizer:
    """
    Builds a new optimizer that generates and evaluates a candidate, finds a neighbor of this candidate,
    and determines if the neighbor is more suitable than the original candidate. It performs this optimization a set
    number of times.

    :param dimensions: dimension of bitstrings passed into fitness_func
    :param restarts: number of times to run the optimization
    :param convergence_iterations: number candidates to evaluate without improvement before declaring convergence
    :param epsilon: minimum required improvement between candidates to continue
    :param fitness_func: function to optimize (accepts a bitstring)
    :param args: will be passed into fitness_func
    :param kwargs: will be passed into fitness_func
    :return: a new optimizer
    """
    random = NumpyRNG()

    random_genome_factory = \
        RandomGenomeFactory(
            dimensions,
            random,
            fitness_func,
            *args,
            **kwargs)

    neighbor_genome_factory = \
        RandomNeighborGenomeFactory(
            dimensions,
            random,
            1,
            fitness_func,
            *args,
            **kwargs)

    restart_convergence = Iterations(restarts)

    single_optimizer_convergence = ConsecutiveNonImprovement(convergence_iterations, epsilon)

    optimizer_builder = \
        BasicOptimizerBuilder(
            dimensions,
            random,
            neighbor_genome_factory,
            single_optimizer_convergence,
            fitness_func,
            *args,
            **kwargs)

    return BasicRestartOptimizer(
        optimizer_builder,
        random_genome_factory,
        1,
        restart_convergence)


def new_simulated_annealer(
        dimensions: Tuple,
        max_neighbor_distance: int,
        initial_temperature: float,
        minimum_temperature: float,
        degradation_multiplier: float,
        fitness_func: Callable[..., float],
        *args,
        **kwargs) -> AbstractOptimizer:
    """
    Creates a new optimizer that optimizes by simulated annealing using a default temperature schedule.

    :param dimensions: dimension of bitstrings passed into fitness_func
    :param max_neighbor_distance: Manhattan distance allowed between neighbors
    :param initial_temperature: starting temperature
    :param minimum_temperature: minimum temperature before convergence
    :param degradation_multiplier: temperature will be multiplied by this every iteration
    :param fitness_func: function to optimize (accepts a bitstring)
    :param args: will be passed into fitness_func
    :param kwargs: will be passed into fitness_func
    :return: a new optimizer
    """
    random = NumpyRNG()

    neighbor_genome_factory = \
        RandomNeighborGenomeFactory(
            dimensions,
            random,
            max_neighbor_distance,
            fitness_func,
            *args,
            **kwargs)

    annealing_schedule = \
        ExponentialAnnealingSchedule(
            initial_temperature,
            minimum_temperature,
            degradation_multiplier)

    initial_candidate = \
        Genome.new_default_genome(
            dimensions,
            fitness_func,
            *args,
            **kwargs)

    initial_candidate.run()

    return BasicAnnealingOptimizer(
        initial_candidate,
        neighbor_genome_factory,
        annealing_schedule,
        random)


def new_hill_climber(
        dimensions: Tuple,
        convergence_iterations: int,
        epsilon: float,
        fitness_func: Callable[..., float],
        *args,
        **kwargs) -> AbstractOptimizer:
    """
    Builds a new optimizer that generates and evaluates a candidate, finds a neighbor of this candidate,
    and determines if the neighbor is more suitable than the original candidate.

    :param dimensions: dimension of bitstrings passed into fitness_func
    :param convergence_iterations: number candidates to evaluate without improvement before declaring convergence
    :param epsilon: minimum required improvement between candidates to continue
    :param fitness_func: function to optimize (accepts a bitstring)
    :param args: will be passed into fitness_func
    :param kwargs: will be passed into fitness_func
    :return: a new optimizer
    """
    random = NumpyRNG()

    neighbor_genome_factory = \
        RandomNeighborGenomeFactory(
            dimensions,
            random,
            1,
            fitness_func,
            *args,
            **kwargs)

    initial_candidate = \
        Genome.new_default_genome(
            dimensions,
            fitness_func,
            *args,
            **kwargs)

    initial_candidate.run()

    convergence_criterion = ConsecutiveNonImprovement(convergence_iterations, epsilon)

    return BasicOptimizer(
        initial_candidate,
        neighbor_genome_factory,
        convergence_criterion)


def new_genetic_optimizer(
        dimensions: Tuple,
        population_size: int,
        mutation_rate: float,
        convergence_iterations: int,
        epsilon: float,
        fitness_func: Callable[..., float],
        *args,
        **kwargs) -> AbstractOptimizer:
    """
    Builds a new optimizer that evaluates a population's fitness, probabilistically chooses the best candidates,
    and constructs a new population from the old. It continues this until the best candidate from the population is
    no longer improving between generations.

    :param dimensions:
    :param population_size: population size of each generation
    :param mutation_rate: probability of a single candidate mutating
    :param convergence_iterations: number candidates to evaluate without improvement before declaring convergence
    :param epsilon: minimum required improvement between candidates to continue
    :param fitness_func: function to optimize (accepts a bitstring)
    :param args: will be passed into fitness_func
    :param kwargs: will be passed into fitness_func
    :return: a new optimizer
    """
    random = NumpyRNG()

    comparer = RouletteWheelComparer(random)

    genome_factory = \
        SinglePointCrossoverGenomeFactory(
            dimensions,
            random,
            mutation_rate,
            comparer,
            fitness_func,
            *args,
            **kwargs)

    initial_population_genome_factory = \
        RandomGenomeFactory(
            dimensions,
            random,
            fitness_func,
            *args,
            **kwargs)

    convergence_criterion = ConsecutiveNonImprovement(convergence_iterations, epsilon)

    initial_population = []

    for _ in range(population_size):
        new_genome = initial_population_genome_factory.build([])
        new_genome.run()
        initial_population.append(new_genome)

    return BasicGeneticOptimizer(
        initial_population,
        genome_factory,
        convergence_criterion)



from typing import List

from tunepy2 import Genome
from tunepy2.interfaces import AbstractGenomeComparer, AbstractRandomNumberGenerator, AbstractLearner
import numpy as np


class RouletteWheelComparer(AbstractGenomeComparer):
    """ Performs a fitness proportionate random model selection. The underlying implementation uses the stochastic
    acceptance method for producing a result in nearly constant time. This is a common method for selecting
    breeding pairs for a genetic algorithm.
    """

    def __init__(self, random_number_generator: AbstractRandomNumberGenerator):
        """ Constructs a new RouletteWheelComparer

        :param random_number_generator: a random number generator
        """
        self._rng = random_number_generator
        self._hash_models = 0
        self._all_fitness = None
        self._max_fitness = -float('inf')

    def compare(self, genomes: List[Genome]) -> Genome:
        """ Compares a list of Genomes and returns a single Genome

        :param genomes: list of Genome objects to compare
        :return: Genome. the result of comparing all genomes
        """

        self.extract_fitness_from_genomes(genomes)

        while True:
            chosen_genome = genomes[self._rng.random_int(0, len(genomes))]
            acceptance_probability = chosen_genome.fitness / self._max_fitness
            if acceptance_probability > self._rng.random():
                return chosen_genome

    def extract_fitness_from_genomes(self, genomes: List[Genome]) -> None:
        new_hash_models = hash(tuple(genomes))

        self._hash_models = new_hash_models
        self._all_fitness = np.zeros(shape=(len(genomes),))

        for index in range(len(genomes)):
            self._all_fitness[index] = genomes[index].fitness

        self._max_fitness = np.max(self._all_fitness)



from abc import ABC, abstractmethod
from typing import Tuple

from tunepy2.interfaces import AbstractLearner


class AbstractValidator(ABC):
    """
    The common interface tunepy2 expects from all validators.
    """

    @abstractmethod
    def validate(self, x, y, model: AbstractLearner) -> float:
        """
        Creates a fitness score for the provided model and data

        :param x: array-like dataset features
        :param y: array-like dataset labels
        :param model: untrained learner
        :return: a fitness score
        """
        pass



from tunepy2.interfaces.abstract_learner import AbstractLearner
from tunepy2.interfaces.abstract_genome_comparer import AbstractGenomeComparer
from tunepy2.interfaces.abstract_optimizer import AbstractOptimizer
from tunepy2.interfaces.abstract_rng import AbstractRandomNumberGenerator
from tunepy2.interfaces.abstract_learner_factory import AbstractLearnerFactory
from tunepy2.interfaces.abstract_validator import AbstractValidator
from tunepy2.interfaces.abstract_genome_factory import AbstractGenomeFactory
from tunepy2.interfaces.abstract_convergence_criterion import AbstractConvergenceCriterion, AbstractAnnealingSchedule
from tunepy2.interfaces.abstract_optimizer_builder import AbstractOptimizerBuilder



import unittest
from tunepy2.validators import TrainTestValidator
from tunepy2.internal import *
from tunepy2.interfaces.stubs import *


class TestTrainTestValidator(unittest.TestCase):
    def test_dimensions_error_on_validate(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TrainTestValidator(0.6)

        with self.assertRaises(DimensionsMismatchException):
            validator.validate(data_features, data_labels, None)

    def test_dimensions_error_on_data_split(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TrainTestValidator(0.6)

        with self.assertRaises(DimensionsMismatchException):
            validator.split_data(data_features, data_labels)

    def test_invalid_split_too_high(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4, 5]

        validator = TrainTestValidator(1.0)

        with self.assertRaises(TrainTestSplitException):
            validator.split_data(data_features, data_labels)

    def test_invalid_split_too_low(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4],
             [5, 5, 5, 5, 5]]

        data_labels = \
            [0, 1, 2, 3, 4, 5]

        validator = TrainTestValidator(0.0)

        with self.assertRaises(TrainTestSplitException):
            validator.split_data(data_features, data_labels)

    def test_bin_sizes(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TrainTestValidator(0.8)
        validator.split_data(data_features, data_labels)

        self.assertEqual(4, len(validator.training_data_labels))
        self.assertEqual(4, len(validator.training_data_features))
        self.assertEqual(1, len(validator.test_data_labels))
        self.assertEqual(1, len(validator.test_data_features))

    def test_validation_result(self):
        data_features = \
            [[0, 0, 0, 0, 0],
             [1, 1, 1, 1, 1],
             [2, 2, 2, 2, 2],
             [3, 3, 3, 3, 3],
             [4, 4, 4, 4, 4]]

        data_labels = \
            [0, 1, 2, 3, 4]

        validator = TrainTestValidator(0.8)
        fitness = validator.validate(data_features, data_labels, ConstantFitnessLearner(69.0))
        self.assertAlmostEqual(69.0, fitness)

if __name__ == '__main__':
    unittest.main()



from tunepy2.optimizers.builders.optimizer_basic_builder import BasicOptimizerBuilder
from tunepy2.optimizers.builders.annealer_basic_builder import BasicAnnealerOptimizerBuilder
from tunepy2.optimizers.builders.genetic_basic_builder import BasicGeneticOptimizerBuilder



from tunepy2.interfaces import AbstractLearnerFactory
from tunepy2.interfaces.stubs import ConstantFitnessLearner


class ConstantFitnessLearnerFactory(AbstractLearnerFactory):
    """
    Builds ConstantFitnessLearners.
    """

    def __init__(self, fitness):
        self._fitness = fitness

    def build(self, bitstring):
        return ConstantFitnessLearner(self._fitness)



import unittest

from tunepy2.interfaces.stubs import PassThroughAnnealingSchedule, PassThroughGenomeFactory, NumpyCustomRandom
from tunepy2.optimizers import BasicAnnealingOptimizer
from tunepy2.internal import Genome


class TestAnnealingOptimizer(unittest.TestCase):
    def test_worse_solution_acceptance(self):
        def fitness_func_zero(bitstring):
            return 0

        def fitness_func_one(bitstring):
            return 1

        initial_candidate = Genome(fitness_func_one, (0,))
        initial_candidate.run()

        annealing_schedule = PassThroughAnnealingSchedule(0.9, True)
        genome_builder = PassThroughGenomeFactory(Genome(fitness_func_zero, (1,)))
        rng = NumpyCustomRandom(0.0, 1)
        optimizer = BasicAnnealingOptimizer(initial_candidate, genome_builder, annealing_schedule, rng)

        optimizer.next()

        self.assertEqual(0, optimizer.best_genome.fitness)
        self.assertAlmostEqual(0.9, optimizer.temperature)

    def test_better_solution_acceptance(self):
        def fitness_func_zero(bitstring):
            return 0

        def fitness_func_one(bitstring):
            return 1

        initial_candidate = Genome(fitness_func_zero, (0,))
        initial_candidate.run()

        annealing_schedule = PassThroughAnnealingSchedule(0.9, True)
        genome_builder = PassThroughGenomeFactory(Genome(fitness_func_one, (1,)))
        rng = NumpyCustomRandom(0.0, 1)
        optimizer = BasicAnnealingOptimizer(initial_candidate, genome_builder, annealing_schedule, rng)

        optimizer.next()

        self.assertEqual(1, optimizer.best_genome.fitness)
        self.assertAlmostEqual(0.9, optimizer.temperature)

    def test_convergence(self):
        def fitness_func_zero(bitstring):
            return 0

        initial_candidate = Genome(fitness_func_zero, (0,))
        initial_candidate.run()

        annealing_schedule = PassThroughAnnealingSchedule(0.9, True)
        genome_builder = PassThroughGenomeFactory(Genome(fitness_func_zero, (1,)))
        rng = NumpyCustomRandom(0.0, 1)
        optimizer = BasicAnnealingOptimizer(initial_candidate, genome_builder, annealing_schedule, rng)

        self.assertFalse(optimizer.converged)
        self.assertAlmostEqual(0.9, optimizer.temperature)
        optimizer.next()
        self.assertTrue(optimizer.converged)
        self.assertAlmostEqual(0.9, optimizer.temperature)


if __name__ == '__main__':
    unittest.main()



import setuptools

with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name="tunepy2",
    version="0.0.2",
    author="Ethan Fortner",
    author_email='ethan.fortner@icloud.com',
    description="A package containing a set of bitstring optimizers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/efortner/tunepy",
    packages=setuptools.find_packages(exclude='test'),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'scikit-learn==0.23.2',
        'numpy==1.19.2',
    ]
)



import unittest
from tunepy2.random import NumpyRNG
import numpy as np


class TestNumpyRNG(unittest.TestCase):
    def test_values_in_range(self):
        random = NumpyRNG(7, integer_type=np.int_)
        array = random.random_int_array(0, 5, shape=(10,))

        for random_number in array:
            self.assertGreaterEqual(random_number, 0)
            self.assertLess(random_number, 5)

    def test_shape_one_dimensional(self):
        random = NumpyRNG(7, integer_type=np.int_)
        array = random.random_int_array(0, 2, shape=(10,))

        self.assertEqual(len(array), 10)
        for random_number in array:
            self.assertIsInstance(random_number, np.int_)

    def test_shape_two_dimensional(self):
        random = NumpyRNG(7, integer_type=np.int_)
        array = random.random_int_array(0, 2, shape=(4, 4))
        number_count = 0

        self.assertEqual(len(array), 4)
        for dimension_two in array:
            self.assertEqual(len(dimension_two), 4)

            for random_number in dimension_two:
                self.assertIsInstance(random_number, np.int_)
                number_count += 1

        self.assertEqual(number_count, 16)

    def test_shape_three_dimensional(self):
        random = NumpyRNG(7, integer_type=np.int_)
        array = random.random_int_array(0, 2, shape=(4, 4, 4))
        number_count = 0

        self.assertEqual(len(array), 4)
        for dimension_two in array:
            self.assertEqual(len(dimension_two), 4)

            for dimension_three in dimension_two:
                self.assertEqual(len(dimension_three), 4)

                for random_number in dimension_three:
                    self.assertIsInstance(random_number, np.int_)
                    number_count += 1

        self.assertEqual(number_count, 64)


if __name__ == '__main__':
    unittest.main()



from tunepy2.interfaces import AbstractValidator


class PassThroughValidator(AbstractValidator):
    """
    A validator that trains one model on all of the provided data and returns its fitness score.
    """

    def validate(self, x, y, model):
        """
        Trains a single model, evaluates on the training set, and returns a fitness score.
        :param x: Array-like dataset features.
        :param y: Array-like dataset labels.
        :param model: An untrained model of a class than implements AbstractLearner.
        :return:
        """

        model.fit(x, y)
        model.evaluate(x)
        return model.fitness



from abc import ABC, abstractmethod
from typing import List

from tunepy2 import Genome


class AbstractGenomeComparer(ABC):
    """
    Interface for comparing multiple Genomes
    """

    @abstractmethod
    def compare(self, genomes: List[Genome]) -> Genome:
        """ Compares a list of Genomes and returns a single Genome

        :param genomes: list of Genome objects to compare
        :return: Genome. the result of comparing all genomes
        """
        pass



from __future__ import annotations

from typing import List, Dict, Callable, Tuple
import numpy as np


class Genome(object):
    """
    Holds a fitness function, a bitstring, and custom arguments.
    """

    def __init__(
            self,
            fitness_func: Callable[..., float],
            bitstring: Tuple,
            *args,
            **kwargs):
        """
        Creates a new Genome.

        :param fitness_func: fitness function passed into new Genome objects that accepts a bitstring
        :param bitstring: will be passed into the fitness function
        :param args: will be passed into the fitness function
        :param kwargs: will be passed into the fitness function
        """
        self._fitness_func = fitness_func
        self._bitstring = bitstring
        self._args = args
        self._kwargs = kwargs
        self._fitness = float("-inf")

    def run(self):
        """
        Executes this Genome's fitness function and updates its fitness.
        """
        self._fitness = self._fitness_func(self._bitstring, *self._args, **self._kwargs)

    @property
    def bitstring(self) -> Tuple:
        """
        This Genome's bitstring

        :return: a tuple representing a bitstring
        """
        return self._bitstring

    @property
    def fitness(self) -> float:
        """
        This Genome's current fitness score

        :return: a float representing fitness
        """
        return self._fitness

    @staticmethod
    def new_default_genome(
            dimensions: Tuple,
            fitness_func: Callable[..., float],
            *args,
            **kwargs) -> Genome:
        """
        Creates a new Genome with a default bitstring

        :param dimensions: tuple containing dimensions of bitstring
        :param fitness_func: fitness function passed into new Genome objects that accepts a bitstring
        :param args: will be passed into the fitness function
        :param kwargs: will be passed into the fitness function
        :return:
        """
        bitstring = np.zeros(shape=dimensions, dtype='int8')
        return Genome(fitness_func, bitstring, *args, **kwargs)



from typing import Tuple

from tunepy2.interfaces import AbstractRandomNumberGenerator
import random


class PythonBaseRNG(AbstractRandomNumberGenerator):
    """
    A tunepy2 wrapper of the built in Python random number generator.
    """

    def __init__(self, seed):
        """
        Creates a new PythonBaseRNG.

        :param seed: seed passed into python's random
        """
        random.seed(seed)

    def random(self) -> float:
        """
        Returns a random number between 0 and 1

        :return: random number between 0 and 1
        """
        return random.random()

    def random_int_array(self, minimum: int, maximum: int, shape: Tuple) -> Tuple:
        """
        Builds an array-like structure of random integers

        :param minimum: minimum value (inclusive)
        :param maximum: maximum value (exclusive)
        :param shape: the shape of the output
        :return: a collection of integers
        """

        def traverse_next_dimension(new_shape, array):
            if len(new_shape) == 1:
                append_new_random_vector(new_shape[0], array)
            else:
                append_new_list_and_recurse(new_shape, array)

        def append_new_random_vector(length, array):
            for _ in range(length):
                array.append(random.randrange(minimum, maximum))

        def append_new_list_and_recurse(new_shape, array):
            for _ in range(new_shape[0]):
                array.append([])
                traverse_next_dimension(new_shape[1:], array[-1])

        return_array = []
        traverse_next_dimension(shape, return_array)
        return tuple(return_array)



from abc import ABC, abstractmethod
from typing import Tuple


class AbstractLearner(ABC):
    """
    The common interface that tunepy2 expects from all machine learning algorithms.
    """

    @abstractmethod
    def fit(self, x_train, y_train):
        """
        Trains the model on a labeled dataset

        :param x_train: table of features
        :param y_train: vector of labels
        """
        pass

    @abstractmethod
    def evaluate(self, x_test, y_test) -> float:
        """
        Returns a fitness score after predicting labels on a dataset

        :param x_test: table of features.
        :param y_test: vector of labels for fitness evaluation.
        :return: A fitness score
        """
        pass

    @property
    @abstractmethod
    def fitness(self) -> float:
        """
        The fitness score of this machine learning model

        :return: The fitness score of this machine learning model
        """
        pass



import unittest

from tunepy2.optimizers.meta import BasicRestartOptimizer
from tunepy2.optimizers.builders import BasicOptimizerBuilder
from tunepy2.interfaces.stubs import PassThroughGenomeFactory, PassThroughConvergenceCriterion
from tunepy2.random import NumpyRNG
from tunepy2 import Genome


class TestOptimizerBasicRestart(unittest.TestCase):
    def test_convergence(self):
        def fitness_func(bitstring):
            return 69.0

        new_population_genome_factory = PassThroughGenomeFactory(
            Genome.new_default_genome(
                (5,),
                fitness_func
            )
        )
        convergence_criterion = PassThroughConvergenceCriterion(True)
        optimizer_builder = BasicOptimizerBuilder(
            (5,),
            NumpyRNG(),
            new_population_genome_factory,
            convergence_criterion,
            fitness_func
        )

        meta_optimizer = BasicRestartOptimizer(
            optimizer_builder,
            new_population_genome_factory,
            10,
            convergence_criterion
        )

        meta_optimizer.next()
        self.assertAlmostEqual(meta_optimizer.best_genome.fitness, 69.0)
        self.assertTrue(meta_optimizer.converged)


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from tunepy2 import transform_bitstring_to_layers_tuple


class TestTransformBitstringToLayersTuple(unittest.TestCase):
    def test_one_dimension_ones(self):
        bitstring = np.ones(shape=(1, 10), dtype=int)
        layers = transform_bitstring_to_layers_tuple(bitstring)
        self.assertEqual(10, layers.shape[0])
        self.assertEqual(10, layers.sum())
        self.assertEqual(0, layers[layers > 1].shape[0])
        self.assertEqual(0, layers[layers < 1].shape[0])

    def test_one_dimension_ones_missing_one(self):
        bitstring = np.ones(shape=(1, 10), dtype=int)
        bitstring[0, 1] = 0
        layers = transform_bitstring_to_layers_tuple(bitstring)
        self.assertEqual(9, layers.shape[0])
        self.assertEqual(9, layers.sum())
        self.assertEqual(0, layers[layers > 1].shape[0])
        self.assertEqual(0, layers[layers < 1].shape[0])

    def test_two_dimension_ones(self):
        bitstring = np.ones(shape=(5, 10), dtype=int)
        layers = transform_bitstring_to_layers_tuple(bitstring)
        self.assertEqual(10, layers.shape[0])
        self.assertEqual(50, layers.sum())
        self.assertEqual(10, layers[layers > 1].shape[0])
        self.assertEqual(0, layers[layers < 1].shape[0])

    def test_two_dimension_ones_missing_one(self):
        bitstring = np.ones(shape=(5, 10), dtype=int)
        bitstring[:, 1] = 0
        layers = transform_bitstring_to_layers_tuple(bitstring)
        self.assertEqual(9, layers.shape[0])
        self.assertEqual(45, layers.sum())
        self.assertEqual(9, layers[layers > 1].shape[0])
        self.assertEqual(0, layers[layers < 1].shape[0])


if __name__ == '__main__':
    unittest.main()



from tunepy2.validators.cross_validator import CrossValidator
from tunepy2.validators.train_test_validator import TrainTestValidator
from tunepy2.validators.time_series_cross_validator import TimeSeriesCrossValidator



from tunepy2.interfaces import AbstractGenomeFactory
from copy import deepcopy


class PassThroughGenomeFactory(AbstractGenomeFactory):
    def __init__(self, returned_genome):
        self._genome = returned_genome
        self._dimensions = ()

    def build(self, prior_genomes):
        return deepcopy(self._genome)

    @property
    def dimensions(self):
        return self._dimensions



from tunepy2 import Genome
from tunepy2.interfaces import AbstractConvergenceCriterion


class Iterations(AbstractConvergenceCriterion):
    """ Declares convergence after a set number of iterations """

    def __init__(self, n: int):
        """ Creates a new Iterations convergence criterion

        :param n: number of iterations before declaring convergence
        """
        self._n = n
        self._current_iteration = 1

    def converged(self, old_candidate: Genome, new_candidate: Genome) -> bool:
        """ Determines if convergence has been reached based on current and prior Genome objects

        :param old_candidate: Genome object from prior iteration
        :param new_candidate: Genome object from current iteration
        :return: true if converged
        """
        if self._current_iteration >= self._n:
            return True

        self._current_iteration += 1
        return False


class ConsecutiveNonImprovement(AbstractConvergenceCriterion):
    """ Declares convergence after a number of iterations without any improvement in fitness """

    def __init__(self, n: int, epsilon: float):
        """ Creates a new ConsecutiveNonImprovement convergence criterion

        :param n: number of iterations
        :param epsilon: minimum required improvement over n iterations
        """
        self._last_iteration = n - 1
        self._iteration_count = 0
        self._epsilon = epsilon
        self._net_improvement = 0.0

    def converged(self, old_candidate: Genome, new_candidate: Genome) -> bool:
        """ Determines if convergence has been reached based on current and prior Genome objects

        :param old_candidate: Genome object from prior iteration
        :param new_candidate: Genome object from current iteration
        :return: true if converged
        """
        if self._net_improvement < self._epsilon and self._iteration_count == self._last_iteration:
            return True

        old_fitness = old_candidate.fitness
        new_fitness = new_candidate.fitness

        if self._iteration_count == 0:
            self._net_improvement = new_fitness - old_fitness
        else:
            self._net_improvement += new_fitness - old_fitness

        if self._net_improvement < self._epsilon:
            self._iteration_count += 1
        else:
            self._iteration_count = 0

        return False
